#include <SPI.h>
#include <Servo.h>
#include <MFRC522.h>
#include <LiquidCrystal.h>

// #define DEBUG

#define MAX_TAGS      100
#define TAG_SIZE        4
#define LOCKED_ANGL   130
#define UNLOCKED_ANGL  34

#define SRV_PIN         8
#define RST_PIN         9
#define SS_PIN         10
#define RS_PIN	        6
#define EN_PIN          7


// Create instances:
// mfrc522 reader
MFRC522 mfrc522(SS_PIN, RST_PIN);
// Parameters: lcd(rs, enable, d2, d3, d4, d5)
LiquidCrystal lcd(RS_PIN, EN_PIN, 2, 3, 4, 5);
// Servo-motor
Servo servo;

String tag_id = "";
int tags_cnt = 0;
char* tags[MAX_TAGS] = { NULL };


void print(String msg, String msg_next_line)
{
#ifndef DEBUG
	lcd.clear();
	lcd.setCursor(0, 0);
      	lcd.print(msg);
	if (msg_next_line != "") {
		lcd.setCursor(0, 1);
      		lcd.print(msg_next_line);
	}
#else
	Serial.print(msg);
	Serial.print("\n");
	if (msg_next_line != "") {
		Serial.print(msg_next_line);
		Serial.print("\n");
	}
#endif
}


void print_normal_mode_msg()
{
	delay(2000);
	print("-Access Control-", "  Scan Your Tag!");
}


bool get_id() {
	// Getting ready for Reading PICCs:

	// If a new PICC placed to RFID reader continue:
	if (!mfrc522.PICC_IsNewCardPresent()) {
		return false;
	}

	// Since a PICC placed get Serial and continue:
	if (!mfrc522.PICC_ReadCardSerial()) {
		return false;
	}

	tag_id = "";
	// The MIFARE PICCs that we use have 4 byte UID:
	for (uint8_t i = 0; i < TAG_SIZE; ++i) {
		// Adds the 4 bytes in a single String variable:
		tag_id.concat(String(mfrc522.uid.uidByte[i], HEX));
	}

	tag_id.toUpperCase();
	mfrc522.PICC_HaltA(); // Stop reading

	return true;
}


void add_or_remove_tag()
{
	// Remove the tag if it exists:
	for (int i = 1; i < tags_cnt; i++) {
		if (tag_id == tags[i]) {
			free(tags[i]);
			tags[i] = tags[--tags_cnt];

			print("  Tag Removed!", "");
	      		print_normal_mode_msg();

	      		return;
	    	}
	}

	// Add the tag if it doesn't exist:
	tags[tags_cnt++] = strdup(tag_id.c_str());

	print("  Tag Added!", "");
	print_normal_mode_msg();
}

void loop_door_closed()
{
	// Checks for tag close by:
	if (!get_id()) {
		return;
	}

    	// Checks whether the scanned tag is the master tag:
    	if (tag_id == tags[0]) {
		print("-Program mode-", "  Add/Remove Tag");

		// Waits until a new card is scanned:
		bool success_read = false;

      		while (!success_read) {
			success_read = get_id();
			
			if (success_read == true) {
	  			add_or_remove_tag();
			}
      		}

		return;
    	}

	// Checks whether the scanned tag is authorized:
	bool correct_tag = false;

    	for (int i = 0; i < 100; i++) {
      		if (tag_id == tags[i]) {
			correct_tag = true;
			print("  Access Granted!", "");

			servo.write(UNLOCKED_ANGL); // Unlocks the door

			print_normal_mode_msg();
			return;
      		}
    	}

    	if (!correct_tag) {
		print("  Access Denied!", "");
      		print_normal_mode_msg();
    	}
}


void loop_door_opened()
{
	print("-Door Opened-", "");

	while (true) {
		int proximitySensor = analogRead(A0);
		if (proximitySensor > 200) {
			break;
		}
	}

	delay(1750);
	servo.write(LOCKED_ANGL); // Locks the door

	print_normal_mode_msg();
}


void setup()
{
	// Initiating:
	SPI.begin();              // SPI bus
  	mfrc522.PCD_Init();       // MFRC522

	Serial.begin(9600);
	lcd.begin(16, 2);         // LCD screen
	servo.attach(SRV_PIN);    // Servo motor
	servo.write(LOCKED_ANGL); // Initial lock position of the servo motor
  
	// Prints the initial message:
	print("-No Master Tag-", "  Scan now!");

	// Waits until a master card is scanned:
	bool success_read = false;

	while (!success_read) {
		success_read = get_id();
		if (success_read == true) {
			// Sets the master tag into position 0 in the array:
			tags[tags_cnt++] = strdup(tag_id.c_str());

			print("  Master Tag Set!", "");
		}
	}

	print_normal_mode_msg();
}

void loop() {
	int proximity_sensor = analogRead(A0);

	if (proximity_sensor > 200) {
		// If door is closed
		loop_door_closed();
	} else {
		// If door is opened
		loop_door_opened();
	}
}
